import asyncio
import json
import logging
import os
import shutil
import signal
# import threading
import time
import zipfile
import zlib

from asyncio_mqtt import Client, MqttError

from spread_core import mqtt
# from spread_core.bam import generator
from spread_core.tools import settings, utils
import spread_lom as lom
from spread_core.bam.validation import enginery, manager, provider, subginery, location, header, cycle, button_preset, trigger, user
import spread_core.bam.validation.constraints as constr
from spread_core.projector.base import *

ZIP_PATH = settings.config[settings.CTP_PATH]
MOSQUITTO_DB = settings.config[settings.MOSQUITTO_PATH]
RECONNECT_INTERVAL = settings.config.get(settings.RECONNECT_INTERVAL, 3)
PROJECT_ID = settings.config[settings.PROJECT_ID]
LOCAL_PATH = os.path.join(settings.config[settings.LOCAL_PATH], mqtt.project.LOCAL)
LOCAL_TOPIC = f"{mqtt.project.ProjectTopic(PROJECT_ID)}/{mqtt.project.LOCAL}"

SERVERS = []         # [server_id]


LOCATIONS = []       # [location_id]
CYCLES = []          # [cycle_id]
PROVIDERS = {}       # {'provider_id': 'provider_type'}
MANAGERS = {}        # {'manager_id': 'manager_type'}
ENGINERIES = {}      # {'enginery_id': 'enginery_type'}
LOCAL_STAGE = True
CTP_ACTUAL = False
ctp = {}
project = {}
start_time = None

# ENG_CONSTR =      {'enginery_type':  {'recipe_type': {'ingredient_type': ['provider_type']}}}
# MAN_PROV_CONSTR = {'manager_type': ['provider_type']}
# SUBG_CONSTR =     {'subginery_type': ['enginery_type']}


class UnsupportedTopic(Exception):
    pass


class ChangeStorage:
    """ Abstract storage of project changes """

    @property
    def topics(self):
        """ Return list of 'Local' subtopics """

        raise NotImplementedError()

    def save(self, topic, data):
        """
        Save topic and data in. If data is None, delete topic.

        Topic is a 'Local' subtopic
        """

        raise NotImplementedError()

    def load(self, topic):
        """
        Return data from topic. Return None, if topic not exist

        Topic is a 'Local' subtopic
        """

        raise NotImplementedError()


class FileSystemStorage(ChangeStorage):
    """ Class for working with File System """

    @property
    def topics(self):
        result = []
        for dirpath, dirnames, filenames in os.walk(LOCAL_PATH):
            for file in filenames:
                topic = f"{dirpath.split(LOCAL_PATH)[1]}/{os.path.splitext(file)[0]}".replace(os.path.sep, '/')[1:]
                result.append(topic)
        return result

    def save(self, topic, data):
        file_dir = os.path.join(LOCAL_PATH, *topic.split('/')[:-1])
        file_path = os.extsep.join((os.path.join(LOCAL_PATH, *topic.split('/')), 'json'))
        if data is None:
            if os.path.exists(file_path):
                os.remove(file_path)
                if not os.listdir(file_dir):
                    os.removedirs(file_dir)
        else:
            if not os.path.exists(file_dir):
                os.makedirs(file_dir, exist_ok=True)
            with open(file_path, mode='w') as file:
                json.dump(data, file, indent=2)

    def load(self, topic):
        file_path = os.extsep.join((os.path.join(LOCAL_PATH, *topic.split('/')), 'json'))
        if os.path.exists(file_path):
            with open(file_path, mode='rb') as file:
                return json.load(file)
        return None


def del_empty_dirs(path):
    for d in os.listdir(path):
        a = os.path.join(path, d)
        if os.path.isdir(a):
            del_empty_dirs(a)
            if not os.listdir(a):
                os.rmdir(a)


fs = FileSystemStorage()


class Launcher:
    """ Abstract launcher """

    def stop(self): raise NotImplemented()

    def update(self): raise NotImplemented()


class Projector(Launcher):
    def stop(self):
        raise KeyboardInterrupt()

    def update(self):
        global LOCAL_STAGE
        LOCAL_STAGE = True
        raise MqttError()


def clear_project(mosquitto_db):
    logging.info('mosquitto clean...')

    os.system('sudo service mosquitto stop')

    if os.path.isdir(mosquitto_db):
        shutil.rmtree(mosquitto_db, ignore_errors=True)
    elif os.path.isfile(mosquitto_db):
        os.remove(mosquitto_db)

    os.system('sudo service mosquitto start')
    logging.info('mosquitto cleaned')


def check_hash(zip_path, mosquitto_db, current_hash):
    real_hash = zlib.crc32(utils.read_file(zip_path, binary=True))

    if real_hash != current_hash:
        clear_project(mosquitto_db)


async def subscribe(client, topic, logged=True):
    await client.subscribe(str(topic))
    if logged:
        logging.debug('subscribed to %s' % topic)


async def unsubscribe(client, topic, logged=True):
    await client.unsubscribe(str(topic))
    if logged:
        logging.debug('unsubscribed from %s' % topic)


async def publish(client, topic: (str, mqtt.Topic), data, retain=True, logged=True):
    if not isinstance(data, (int, str, bytes, bytearray)):
        data = json.dumps(data)

    if logged:
        logging.debug('[%s] -> %s' % (topic, data if data.__sizeof__() <= 1024 else ('%s bytes' % len(str(data)))))
    await client.publish(str(topic), data, retain=retain)


async def hash_ctp_publish(client):
    # logging.debug('hash_ctp_published started')
    try:
        await asyncio.sleep(1)  # waiting for Hash message
    except asyncio.CancelledError:
        pass
    else:
        zip_data = utils.read_file(ZIP_PATH, binary=True)
        await client.publish(str(mqtt.project.ProjectFile(PROJECT_ID, mqtt.project.CTP)), payload=zip_data, retain=True)
        await client.publish(str(mqtt.project.Hash(PROJECT_ID)), payload=zlib.crc32(zip_data), retain=True)
        logging.debug('hash and ctp published')


async def project_publish(client):
    for topic in project:
        obj = project[topic]
        await client.publish('/'.join((LOCAL_TOPIC, topic)), payload=json.dumps(obj), retain=True)
    if not project:
        logging.debug("project don't need publish")
    else:
        logging.debug('project published')
    project.clear()


async def local_subscribe(client):
    try:
        await asyncio.sleep(1)
    except asyncio.CancelledError:
        pass
    else:
        await unsubscribe(client, mqtt.project.Hash(PROJECT_ID))
        await subscribe(client, mqtt.project.Local.broadcast(PROJECT_ID, '#'))


async def local_unsubscribe(client):
    global LOCAL_STAGE
    try:
        await asyncio.sleep(1)
    except asyncio.CancelledError:
        pass
    else:
        logging.debug(f"Broker check time: {time.time()-start_time}")
        await unsubscribe(client, mqtt.project.Local.broadcast(PROJECT_ID, '#'))
        LOCAL_STAGE = False
        # os.system('sudo spread_utils_dashboard fill --all')  ???
        await project_publish(client)
        await subscribe(client, mqtt.project.Set.broadcast(PROJECT_ID, '#'))
        await subscribe(client, mqtt.project.Reset.broadcast(PROJECT_ID, '#'))


async def project_init(client):
    """
        - Создаем ctp в том виде, как он должен быть опубликован {топик: значение}
        - project = ctp + изменения из файловой
    """
    global fs, ctp, project, start_time, PROJECT_ID, SERVERS, LOCATIONS, CYCLES, ENGINERIES, PROVIDERS, MANAGERS

    ERRORS = False
    start_time = time.time()

    with zipfile.ZipFile(ZIP_PATH, 'r') as input_zip:
        with input_zip.open(f"{mqtt.project.ProjectFile.HEADER}.json", "r") as file:
            header_file = json.load(file)['header']

        try:
            header.HeaderValidator(header_file)
        except ValueError as ex:
            raise ValueError(f"Init error: Wrong {mqtt.project.ProjectFile.HEADER}.json ({ex})")

        if PROJECT_ID != header_file['project'].pop('id'):
            raise ValueError('Init error: projectID in config file not corresponding of ctp project!')

        topic = f"Header/Project"
        ctp[topic] = header_file['project']

        topic = f"Header/Firmware"
        ctp[topic] = header_file['firmware']

        del header_file

        with input_zip.open(f"{mqtt.project.ProjectFile.SERVERS}.json", "r") as file:
            servers = json.load(file)['servers']

        SERVERS = [obj['id'] for obj in servers]

        del servers

        with input_zip.open(f"{mqtt.project.ProjectFile.MANAGERS}.json", "r") as file:
            managers = json.load(file)['managers']

        MANAGERS = {}
        for obj in managers:
            obj_id = obj.pop('id')
            if obj_id not in MANAGERS:
                MANAGERS[obj_id] = obj['type']
            else:
                ERRORS = True
                logging.debug(f"Init error: Wrong manager (id={obj_id}, id not unique)")


            topic = f"Hardware/{obj.pop('type')}/{obj_id}"
            if data_valid(LOCAL_TOPIC + '/' + topic, obj):
                ctp[topic] = obj
            else:
                ERRORS = True
                logging.debug(f"Init error: Wrong manager (id={obj_id})")

        del managers

        with input_zip.open(f"{mqtt.project.ProjectFile.PROVIDERS}.json", "r") as file:
            providers = json.load(file)['providers']

        PROVIDERS = {}
        for obj in providers:
            obj_id = obj.pop('id')
            if obj_id not in PROVIDERS:
                PROVIDERS[obj_id] = obj['type']
            else:
                ERRORS = True
                logging.debug(f"Init error: Wrong provider (id={obj_id}, id not unique)")

            topic = f"Hardware/{MANAGERS[obj['managerID']]}/{obj.pop('managerID')}/{obj.pop('type')}/{obj_id}"
            if data_valid(LOCAL_TOPIC + '/' + topic, obj):
                ctp[topic] = obj
            else:
                ERRORS = True
                logging.debug(f"Init error: Wrong provider (id={obj_id})")

        del providers

        with input_zip.open(f"{mqtt.project.ProjectFile.ENGINERIES}.json", "r") as file:
            engineries = json.load(file)['engineries']

        ENGINERIES = {}
        for obj in engineries:
            obj_id = obj.pop('id')
            if obj_id not in ENGINERIES:
                ENGINERIES[obj_id] = obj['type']
            else:
                ERRORS = True
                logging.debug(f"Init error: Wrong enginery (id={obj_id}) (id not unique)")

            topic = f"Equipment/{obj.pop('type')}/{obj_id}"
            if data_valid(LOCAL_TOPIC + '/' + topic, obj):
                ctp[topic] = obj
            else:
                ERRORS = True
                logging.debug(f"Init error: Wrong enginery (id={obj_id})")

        del engineries

        with input_zip.open(f"{mqtt.project.ProjectFile.LOCATIONS}.json", "r") as file:
            locations = json.load(file)['locations']

        LOCATIONS = []
        for obj in locations:
            obj_id = obj.pop('id')
            if obj_id not in LOCATIONS:
                LOCATIONS.append(obj_id)
            else:
                ERRORS = True
                logging.debug(f"Init error: Wrong location (id={obj_id}, id not unique)")

            topic = f"Location/{obj_id}"
            if data_valid(LOCAL_TOPIC + '/' + topic, obj):
                ctp[topic] = obj
            else:
                ERRORS = True
                logging.debug(f"Init error: Wrong location (id={obj_id})")

        del locations

        with input_zip.open(f"{mqtt.project.ProjectFile.SUBGINERIES}.json", "r") as file:
            subgineries = json.load(file)['subgineries']

        SUBGINERIES = []
        for obj in subgineries:
            obj_id = obj.pop('id')
            if obj_id not in SUBGINERIES:
                if obj_id not in ENGINERIES:
                    SUBGINERIES.append(obj_id)
                else:
                    ERRORS = True
                    logging.debug(f"Init error: Wrong subginery (id={obj_id}, id not unique, enginery got the same id)")
            else:
                ERRORS = True
                logging.debug(f"Init error: Wrong subginery (id={obj_id}, id not unique)")

            topic = f"Location/{obj.pop('locationID')}/{obj.pop('type')}/{obj_id}"
            if data_valid(LOCAL_TOPIC + '/' + topic, obj):
                ctp[topic] = obj
            else:
                ERRORS = True
                logging.debug(f"Init error: Wrong subginery (id={obj_id})")

        del subgineries, SUBGINERIES

        # users.json пока нет в ctp !!!
        users_file = mqtt.project.ProjectFile.USERS + '.json'
        if users_file in input_zip.namelist():
            with input_zip.open(users_file, "r") as file:  # , encoding="utf-8" ???
                users = json.load(file)['users']

            USERS = []
            for obj in users:
                obj_id = obj.pop('id')
                if obj_id not in USERS:
                    USERS.append(obj_id)
                else:
                    ERRORS = True
                    logging.debug(f"Init error: Wrong user (id={obj_id}, id not unique)")

                topic = f"Users/{obj_id}"
                if data_valid(LOCAL_TOPIC + '/' + topic, obj):
                    ctp[topic] = obj
                else:
                    ERRORS = True
                    logging.debug(f"Init error: Wrong user (id={obj_id})")

            del users, USERS

    logging.debug(f"ctp processing time: {time.time()-start_time}")

    # FS modify
    for topic in fs.topics:
        try:
            await change_storage_modify(topic, client, LOCAL_TOPIC)
        except UnsupportedTopic as ex:
            # logging.info(str(ex))
            pass

    # FS apply to project
    CYCLES = []
    project = ctp.copy()
    for topic in fs.topics:
        if topic in project or topic.startswith(mqtt.project.Preset.family) \
                or topic.startswith(mqtt.project.Cycle.family) or topic.startswith(mqtt.project.User.family):
            try:
                data = fs.load(topic)
            except:
                ERRORS = True
                logging.debug(f"Init error: Wrong FS entity structure (topic='{topic}')")
                continue

            # Нужно, пока в FS Triggers хранятся в Cycle
            # !!! Если триггеры будут храниться отдельно,
            # !!! будет невозможно контролировать последовательность топиков Cycle - Triggers
            if topic.startswith(mqtt.project.CYCLE):
                try:
                    topic_obj = mqtt.topic_of('/'.join((LOCAL_TOPIC, topic)))
                except:
                    ERRORS = True
                    logging.debug(f"Init error: Wrong FS entity (topic='{topic}')")
                    continue

                CYCLES.append(topic_obj.entity_id)
                if 'triggers' in data:
                    triggers = data.pop('triggers')
                    for tr_id in triggers:
                        tr_topic = '/'.join((topic, mqtt.project.TRIGGERS, tr_id))
                        if data_valid(LOCAL_TOPIC + '/' + tr_topic, triggers[tr_id]):
                            project[tr_topic] = triggers[tr_id]

            if data_valid(LOCAL_TOPIC + '/' + topic, data):
                project[topic] = data
            else:
                ERRORS = True
                logging.debug(f"Init error: Wrong FS entity (topic='{topic}')")
        else:
            ERRORS = True
            logging.debug(f"Init error: Wrong FS entity (topic='{topic}')")

    if ERRORS:
        raise ValueError('Init error!')

    logging.debug('project ready to publish')


def topic_valid(topic) -> bool:
    try:
        mqtt.topic_of(topic)
        return True
    except:
        return False


def data_valid(topic, data) -> bool:
    try:
        topic_object = mqtt.topic_of(topic)

        if isinstance(topic_object, mqtt.project.Equipment):  # Enginery
            # generator.generate_enginery(topic_object.entity_id, topic_object.entity_type, data)
            enginery.EngineryValidator({'id': topic_object.entity_id, 'type': topic_object.entity_type, **data})

            # recipe ingredient id must be real provider_id and provider_type must be in Constraints
            if 'recipe' in data:
                if 'ingredients' in data['recipe']:
                    for ingr in data['recipe']['ingredients']:
                        if ingr['id'] not in PROVIDERS:
                            raise ValueError('Wrong recipe ingredient id (not in project)')

                        if ingr['type'] in constr.ENG_CONSTR[topic_object.entity_type][data['recipe']['type']]:
                            ingr_type = ingr['type']
                        elif ingr['type'] + '[]' in constr.ENG_CONSTR[topic_object.entity_type][data['recipe']['type']]:
                            ingr_type = ingr['type'] + '[]'
                        else:
                            raise ValueError('Wrong recipe ingredient id type (not in constraints)')

                        if PROVIDERS[ingr['id']] not in constr.ENG_CONSTR[topic_object.entity_type][data['recipe']['type']][ingr_type]:
                            raise ValueError('Wrong recipe ingredient id type (not in constraints)')

        elif isinstance(topic_object, mqtt.project.Hardware):  # Manager
            manager.ManagerValidator({'id': topic_object.entity_id, 'type': topic_object.entity_type, **data})
            if data['serverID'] not in SERVERS:
                raise ValueError('Wrong serverID (not in project)')

        elif isinstance(topic_object, mqtt.project.Provider):
            # generator.generate_provider(topic_object.entity_id, topic_object.entity_type, data)
            provider.ProviderValidator({'id': topic_object.entity_id, 'type': topic_object.entity_type, 'managerID': topic_object.manager_id, **data})

            # managerID must be real and manager_type must be in Constraints
            if topic_object.manager_id not in MANAGERS:
                raise ValueError('Wrong managerID (not in project)')
            if topic_object.manager_type != MANAGERS[topic_object.manager_id]:
                raise ValueError('Wrong managerID type')
            if topic_object.manager_type not in constr.MAN_PROV_CONSTR:
                raise ValueError('Wrong managerID type (not in constraints)')
            if topic_object.entity_type not in constr.MAN_PROV_CONSTR[topic_object.manager_type]:
                raise ValueError('Wrong provider type (not in constraints)')

        elif isinstance(topic_object, mqtt.project.Subginery):
            # generator.generate_subginery(topic_object.entity_id, topic_object.entity_type, data)
            subginery.SubgineryValidator({'id': topic_object.entity_id, 'type': topic_object.entity_type, 'locationID': topic_object.location_id, **data})

            # engineries must be real and enginery_type must be in Constraints
            if 'engineries' in data:
                for eng_id in data['engineries']:
                    if eng_id not in ENGINERIES:
                        raise ValueError('Wrong engineries (not in project)')
                    if ENGINERIES[eng_id] not in constr.SUBG_CONSTR[topic_object.entity_type]:
                        raise ValueError('Wrong engineries (not in constraints)')

        elif isinstance(topic_object, mqtt.project.Location):
            location.LocationValidator({'id': topic_object.entity_id, **data})
            if topic_object.entity_id not in LOCATIONS:
                raise ValueError('Wrong id')

        elif isinstance(topic_object, mqtt.project.Preset):
            if topic_object.entity_type == 'ButtonPresets':
                button_preset.ButtonPresetValidator({'id': topic_object.entity_id, **data})

            elif topic_object.entity_type == 'LomProfiles':
                lom.profiles.GlobalDimmingProfile.deserialize({'id': topic_object.entity_id, **data})

        elif isinstance(topic_object, mqtt.project.Cycle):
            cycle.CycleValidator({'id': topic_object.entity_id, **data})

        elif isinstance(topic_object, mqtt.project.CycleTrigger):
            if topic_object.cycle_id not in CYCLES:
                raise ValueError('Wrong cycle id')
            trigger.TriggerValidator({'id': topic_object.entity_id, **data})

        elif isinstance(topic_object, mqtt.project.User):
            user.UserValidator({'id': topic_object.entity_id, **data})
            if data['rootLocationID'] not in LOCATIONS:
                raise ValueError('Wrong rootLocationID')

            if 'homeLocationID' in data and data['homeLocationID'] not in LOCATIONS:
                raise ValueError('Wrong homeLocationID')
        else:
            return False

        return True
    except ValueError or BaseException as ex:
        logging.debug('Validation error: %s' % ex)
        return False


async def listener():
    while True:
        try:
            async with Client(
                    hostname=settings.config[settings.BROKER].get(settings.HOST, '127.0.0.1'),
                    port=settings.config[settings.BROKER].get(settings.PORT, 1883),
                    username=settings.config[settings.BROKER].get(settings.LOGIN),
                    password=settings.config[settings.BROKER].get(settings.PASSWORD),
                    client_id='Projector'
            ) as client:

                async with client.filtered_messages(str(mqtt.project.ProjectTopic.broadcast(PROJECT_ID, '#'))) as messages:  # "Project/id/#"

                    if LOCAL_STAGE:  # При старте сервиса True
                        try:
                            await project_init(client)
                        except ValueError as ex:
                            logging.exception(ex)
                            break

                        ctp_hash = zlib.crc32(utils.read_file(ZIP_PATH, binary=True))
                        # print(mqtt.project.Local.broadcast(PROJECT_ID, '#'))
                        # raise KeyboardInterrupt
                        await subscribe(client, mqtt.project.Hash(PROJECT_ID))                  # "Project/id/Hash"
                        await subscribe(client, mqtt.project.Local.broadcast(PROJECT_ID, '#'))  # "Project/id/Local/#"
                        hash_ctp_publish_task = asyncio.create_task(hash_ctp_publish(client))
                        local_unsubscribe_task = asyncio.create_task(local_unsubscribe(client))
                        global start_time
                        start_time = time.time()
                    else:
                        await subscribe(client, mqtt.project.Set.broadcast(PROJECT_ID, '#'))    # "Project/id/Set/#"
                        await subscribe(client, mqtt.project.Reset.broadcast(PROJECT_ID, '#'))  # "Project/id/Reset/#"

                    async for message in messages:
                        try:
                            if not message.payload:
                                continue

                            topic = message.topic
                            data = json.loads(message.payload)

                            if LOCAL_STAGE:
                                if local_unsubscribe_task:
                                    local_unsubscribe_task.cancel()
                                    local_unsubscribe_task = None

                                if topic == mqtt.project.Hash(PROJECT_ID):
                                    if data == ctp_hash:
                                        if hash_ctp_publish_task:
                                            hash_ctp_publish_task.cancel()
                                            hash_ctp_publish_task = None
                                            logging.debug("hash and ctp don't need publish")
                                    await unsubscribe(client, mqtt.project.Hash(PROJECT_ID))

                                elif topic.startswith(f"{LOCAL_TOPIC}/"):
                                    topic = topic.replace(f"{LOCAL_TOPIC}/", '', 1)
                                    if topic in project:
                                        if data == project[topic]:
                                            del project[topic]

                                    elif topic_valid(message.topic):
                                        await client.publish(message.topic, payload=bytes(), retain=True)

                                local_unsubscribe_task = asyncio.create_task(local_unsubscribe(client))
                                continue

                            # "Project/id/Set/#" messages analytical block
                            elif topic.startswith(f"{mqtt.project.Set(PROJECT_ID)}/"):
                                topic = topic.replace(f"{mqtt.project.Set(PROJECT_ID)}/", '', 1)
                                if topic in ctp:
                                    fs_data = fs.load(topic)
                                    if data == ctp[topic]:
                                        if fs_data:
                                            fs.save(topic, None)
                                            await publish(client, '/'.join((LOCAL_TOPIC, topic)), ctp[topic])
                                    elif data != fs_data:
                                        if data_valid('/'.join((LOCAL_TOPIC, topic)), data):
                                            if (fs_data and data['key'] == fs_data['key']) or data['key'] == ctp[topic]['key']:
                                                raise ValueError('Set error: New key needed')
                                            fs.save(topic, data)
                                            await publish(client, '/'.join((LOCAL_TOPIC, topic)), data)
                                    else:
                                        raise ValueError('Set error: No data change in topic: "%s"' % topic)
                                    continue

                                try:
                                    topic_object = mqtt.topic_of('/'.join((LOCAL_TOPIC, topic)))
                                except:
                                    raise ValueError('Set error: Invalid topic: "%s"' % topic)

                                if isinstance(topic_object, mqtt.project.Preset) or isinstance(topic_object, mqtt.project.Cycle):
                                    fs_data = fs.load(topic)
                                    if data != fs_data:
                                        if data_valid('/'.join((LOCAL_TOPIC, topic)), data):
                                            fs.save(topic, data)
                                            await publish(client, '/'.join((LOCAL_TOPIC, topic)), data)
                                            if isinstance(topic_object, mqtt.project.Cycle) and topic_object.entity_id not in CYCLES:
                                                CYCLES.append(topic_object.entity_id)
                                        else:
                                            raise ValueError('Set error: Invalid data of topic: "%s"' % topic)
                                    else:
                                        raise ValueError('Set error: No data change in topic: "%s"' % topic)

                                elif isinstance(topic_object, mqtt.project.CycleTrigger):
                                    # В FS Triggers хранятся в Cycle !!!
                                    if data_valid('/'.join((LOCAL_TOPIC, topic)), data):
                                        cycle_topic = '/'.join((mqtt.project.CYCLE, str(topic_object.cycle_id)))
                                        cycle_data = fs.load(cycle_topic)
                                        triggers = cycle_data.get('triggers', None)
                                        trig_id = str(topic_object.entity_id)

                                        if triggers and trig_id in triggers and data == triggers[trig_id]:
                                            raise ValueError('Set error: No data change in topic: "%s"' % topic)

                                        if not triggers:
                                            cycle_data['triggers'] = {}

                                        cycle_data['triggers'][trig_id] = data
                                        fs.save(cycle_topic, cycle_data)
                                        await publish(client, '/'.join((LOCAL_TOPIC, topic)), data)
                                    else:
                                        raise ValueError('Set error: Invalid data of topic: "%s"' % topic)
                                else:
                                    raise ValueError('Set error: Invalid topic: "%s"' % topic)

                            # "Project/id/Reset/#" messages analytical block
                            elif topic.startswith(f"{mqtt.project.Reset(PROJECT_ID)}/"):
                                topic = topic.replace(f"{mqtt.project.Reset(PROJECT_ID)}/", '', 1)
                                if not isinstance(data, dict):
                                    raise ValueError('Reset error: Invalid data type of topic: "%s"' % topic)

                                if topic in ctp:
                                    fs.save(topic, None)
                                    await publish(client, '/'.join((LOCAL_TOPIC, topic)), ctp[topic])
                                    logging.debug(f'Reset topic: "%s"' % topic)
                                    continue

                                try:
                                    topic_object = mqtt.topic_of('/'.join((LOCAL_TOPIC, topic)))
                                except:
                                    fs.save(topic, None)
                                    await client.publish('/'.join((LOCAL_TOPIC, topic)), payload=bytes(), retain=True)
                                    logging.debug(f'Reset topic: "%s"' % topic)
                                    continue
                                    # raise ValueError('Reset error: Invalid topic "%s"' % topic)

                                if isinstance(topic_object, mqtt.project.Cycle):
                                    # При удалении Cycle, удаляем и Triggers из брокера
                                    fs_data = fs.load(topic)
                                    if fs_data:
                                        if 'triggers' in fs_data:
                                            for tr_id in fs_data['triggers']:
                                                await client.publish('/'.join((LOCAL_TOPIC, topic, mqtt.project.TRIGGERS, tr_id)), payload=bytes(), retain=True)

                                        fs.save(topic, None)
                                        CYCLES.remove(topic_object.entity_id)

                                    await client.publish('/'.join((LOCAL_TOPIC, topic)), payload=bytes(), retain=True)
                                    # else:
                                    #     raise ValueError('Reset error: Invalid topic "%s"' % topic)

                                elif isinstance(topic_object, mqtt.project.CycleTrigger):
                                    # В FS Triggers хранятся в Cycle !!!
                                    cycle_topic = '/'.join((mqtt.project.CYCLE, str(topic_object.cycle_id)))
                                    cycle_data = fs.load(cycle_topic)
                                    if not cycle_data:
                                        await client.publish('/'.join((LOCAL_TOPIC, topic)), payload=bytes(), retain=True)
                                        logging.debug(f'Reset topic: "%s"' % topic)
                                        continue
                                        # raise ValueError('Reset error: Invalid topic "%s"' % topic)

                                    triggers = cycle_data.get('triggers', None)
                                    trig_id = str(topic_object.entity_id)

                                    if triggers and trig_id in triggers:
                                        del cycle_data['triggers'][trig_id]
                                        if not cycle_data['triggers']:
                                            del cycle_data['triggers']
                                        fs.save(cycle_topic, cycle_data)

                                    await client.publish('/'.join((LOCAL_TOPIC, topic)), payload=bytes(), retain=True)
                                    # else:
                                    #     raise ValueError('Reset error: Invalid topic "%s"' % topic)

                                else:
                                    fs.save(topic, None)
                                    await client.publish('/'.join((LOCAL_TOPIC, topic)), payload=bytes(), retain=True)

                                logging.debug(f'Reset topic: "%s"' % topic)

                        except BaseException as ex:
                            logging.exception('%s' % ex)
        except MqttError as error:
            logging.error(error)
            await asyncio.sleep(RECONNECT_INTERVAL)
        except (KeyboardInterrupt, asyncio.CancelledError):
            print('breaked')
            break
        except BaseException as ex:
            logging.exception('%s' % ex)


async def change_storage_modify(topic, client, local_topic):
    topic_mod = TopicModifier(fs, client, local_topic)

    if topic.startswith('Equipment/Lighting/LightingArea/ButtonPresets'):
        new_topic = topic.replace('Equipment/Lighting/LightingArea', mqtt.project.Preset.family, 1)
        topic_mod.modify(topic, new_topic)
        # fs.save(new_topic, fs.load(topic))
        # fs.save(topic, None)
        # await client.publish('/'.join((LOCAL_TOPIC, topic)), payload=bytes(), retain=True)

    elif topic.startswith('Hardware/Lom/Profiles'):
        new_topic = topic.replace('Hardware/Lom/Profiles', f'{mqtt.project.Preset.family}/LomProfiles', 1)
        fs.save(new_topic, fs.load(topic))
        fs.save(topic, None)
        await client.publish('/'.join((LOCAL_TOPIC, topic)), payload=bytes(), retain=True)

    elif topic.startswith('Incubator/Cycles'):
        new_topic = topic.replace('Incubator/Cycles', mqtt.project.Cycle.family, 1)
        fs.save(new_topic, fs.load(topic))
        fs.save(topic, None)
        await client.publish('/'.join((LOCAL_TOPIC, topic)), payload=bytes(), retain=True)
    else:
        raise UnsupportedTopic("Unsupported topic: %s" % topic)


def run():
    # projector = Projector()

    # signal.signal(signal.SIGHUP, lambda signum, frame: settings.config.parse())
    try:
        asyncio.run(listener())
    except KeyboardInterrupt:
        pass


if __name__ == '__main__':
    run()
