from unittest import TestCase

from spread_core.bam.engineries import Enginery

from spread_core.bam.engineries.climate import TemperatureSensor, Thermoregulator, Fan, HeatedFloor

from spread_core.bam.subgineries import climating as climating
from spread_core.mqtt.spread.variable import Variable
from spread_core.tests.equipment.utils import TestEquipmentAdapter, TestEngineryRecipe


class TestClimate(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.adapter = TestEquipmentAdapter()

    def setUp(self) -> None:
        ENGS = [TemperatureSensor, Thermoregulator, Fan, HeatedFloor, Enginery]

        self.adapter.add_manager(2, "Manager")

        for i in range(15):
            e = ENGS[i//3](i)
            self.adapter.add_provider(TestEngineryRecipe(1000 + i))
            self.adapter.add_enginery(e, 1000+i, "Simple")

        self._sb = climating.Climate(99, 2, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14])
        self._sb.set_adapter(self.adapter)

    def _check_publish_state(self, funit, value, delay=0, params=None):
        self.__check_publish(f'Spread/State/1/Location/99/Climate/{funit}', value, _retain=True, _delay=delay, params=params)

    def _check_publish_command(self, p_id, funit, value, params=None):
        self.__check_publish(f'Spread/Set/Provider/{p_id}/{funit}', value, _retain=False, _delay=0, params=params)

    def _check_end_of_publishes(self):
        if self.adapter.queue:
            unexpected_publishes = f'Unexpected publishes({len(self.adapter.queue)} pcs):\n'
            unexpected_publishes += "\n".join(str(i) for i in self.adapter.queue)
            self.fail(unexpected_publishes)

    def __check_publish(self, _topic, _value, _retain=True, _delay=0, params=None):
        if len(self.adapter.queue) == 0:
            self.fail('Expected publish is absent: %s' % _topic)
        else:
            topic, variable, retain, delay = self.adapter.queue.popleft()
            self.assertEqual(_topic, topic)
            self.assertEqual(_value, variable.value)
            self.assertEqual(_retain, retain)
            self.assertEqual(_delay, delay)

            if params:
                for param_name in params:
                    try:
                        self.assertEqual(params[param_name], variable.params[param_name])
                    except KeyError:
                        self.fail('Expected variable parameter "%s" is absent' % param_name)

    @property
    def sb(self) -> climating.Climate:
        return self._sb

    def test_init(self):
        self.assertEqual(1, self.sb.project_id)
        self.assertEqual(2, self.sb.location_id)
        self.assertEqual(99, self.sb.id)
        self.assertEqual(self.sb._enginery_ids, [e.id for e in self.sb.engineries])

        for e in self.sb.engineries:
            assert isinstance(e, Enginery)
            self.assertIn(self.sb, e._subgineries)

    def test_address(self):
        self.assertEqual('1/Location/99/Climate/SomeFunit', self.sb.get_address("SomeFunit"))

    def test_state_publishing(self):
        self.sb.publish_state('SomeFunit', 'SomeValue')
        self._check_publish_state('SomeFunit', 'SomeValue')
        self._check_end_of_publishes()

    def test_invalidation(self):
        self.sb.values['SomeFunit'] = 2
        self.sb.set_invalid()

        self._check_publish_state('SomeFunit', None)
        self._check_end_of_publishes()

    def _set_enginery_temperature(self, enginery_id, value, expected_unit=None, expected_value=None):
        e = self.adapter._entities[enginery_id]
        assert isinstance(e, Enginery)
        e.values[expected_unit] = value
        self.sb.on_level_changed(e)
        if expected_unit:
            self._check_publish_state(expected_unit, expected_value)

        self._check_end_of_publishes()

    def test_fire_changing(self):
        self._set_enginery_temperature(0, 29815, 'Temperature', 29815)
        self._set_enginery_temperature(1, 29615, 'Temperature', 29715)
        self._set_enginery_temperature(2, 29415, 'Temperature', 29615)
        self._set_enginery_temperature(0, None, 'Temperature', 29515)
        self._set_enginery_temperature(1, None, 'Temperature', 29415)
        self._set_enginery_temperature(2, None, 'Temperature', None)
